/* File: Magnify.m - Application class for 'Magnify'
 *
 * By: Christopher Lane <lane@camis.stanford.edu>
 * Symbolic Systems Resources Group
 * Knowledge Systems Laboratory
 * Stanford University
 *
 * PostScript optimizations by: Scott Byer <byer@adobe.com>
 *
 * Date:  13 May 1994
 *
 * Copyright: 1990, 1991, 1992  & 1994 by The Leland Stanford Junior University.
 * This program may be distributed without restriction for non-commercial use.
 */

#import <ctype.h>
#import <appkit/appkit.h>

#import "Magnify.h"
#import "MagnifyView.h"
#import "DefaultsTable.h"
#import "PopUpListPatch.h"

#define VERSION __DATE__

#define FROZENMENUCELLTAG (1)
#define FROZENMENUTITLE "Freeze"
#define UNFROZENMENUTITLE "Unfreeze"

#define WINDOWNAME "MagnifyWindow"
#define WINDOWTITLEFORMAT "%.2gx Magnification"
#define WINDOWTITLEPROTOTYPE "0.00x Magnification"

#define SEPARATOR ":"
#define DEFAULTSFILE "Defaults"
#define ORIGINALDEFAULTSET "Original"

typedef enum {NONE, CURRENT, CROSS} CURSORS;

#define streq(s, t) (strcmp(s, t) == 0)

#define originalKey(n) keycat(NULL, n)
#define validDefault(s, n) (getDefault(keycat(s, n)) != NULL)

#define getDefaultSet(s, n) ((validDefault(s, n)) ? getDefault(keycat(s, n)) : getDefault(originalKey(n)))
#define getIntDefaultSet(s, n) ((validDefault(s, n)) ? getIntDefault(keycat(s, n)) : getIntDefault(originalKey(n)))
#define getFloatDefaultSet(s, n) ((validDefault(s, n)) ? getFloatDefault(keycat(s, n)) : getFloatDefault(originalKey(n)))
#define getBoolDefaultSet(s, n) ((validDefault(s, n)) ? getBoolDefault(keycat(s, n)) : getBoolDefault(originalKey(n)))

#define writeDefaultSet(s, n, v) writeDefault(keycat(s, n), v)
#define writeIntDefaultSet(s, n, v) writeIntDefault(keycat(s, n), v)
#define writeFloatDefaultSet(s, n, v) writeFloatDefault(keycat(s, n), v)
#define writeBoolDefaultSet(s, n, v) writeBoolDefault(keycat(s, n), v)

#define removeDefaultSet(s, n) removeDefault(keycat(s, n))

const char *keycat(const char *s, const char *n)
{
	static char buffer[BUFSIZ];
	
	if (s == NULL) return n;
	
	return strcat(strcat(strcpy(buffer, s), " "), n);
}

@implementation Magnify : Application

+ new
{
	NXBundle *bundle;
	char pathnamebuf[MAXPATHLEN];

	self = [super new];

	bundle = [NXBundle bundleForClass:[self class]];
	
	if ([bundle getPath:pathnamebuf forResource:DEFAULTSFILE ofType:"strings"])

	[(defaults = [[DefaultsTable alloc] initFromFile:pathnamebuf]) registerDefaults:[self appName]];

	return self;
}

- appDidInit:sender
{	
	MenuCell *cell;
	NXPoint point = {7, -7};
	Menu *menu = [self mainMenu];

	[[(crossCursor = [NXCursor alloc]) initFromImage:[NXImage findImageNamed:"Cross"]] setHotSpot:&point];

	[self loadDefaultSets];

	[magnifyWindow makeFirstResponder:[magnifyView setNextResponder:[magnifyWindow firstResponder]]];

	(void) [magnifyWindow addToEventMask:(NX_FLAGSCHANGED)];

	if ((cell = [menu findCellWithTag:FROZENMENUCELLTAG]) != nil) {
		[cell setUpdateAction:@selector(isFrozenUpdateAction:) forMenu:menu];
		[menu update];
		}
		
	[[magnifyView updateImage] display];
	
	[magnifyWindow makeKeyAndOrderFront:self];
	
	return [self setDelegate:self];
}

- appWillTerminate:sender
{
	return [self saveDefaultSetNamed:[defaultField stringValue]];
}

- appDidHide:sender
{
	[magnifyView freeze];
	
	return self;
}

- appDidUnhide:sender
{
	[magnifyView unfreeze];
	
	return self;
}

- adjust:sender
{
	id slider;
	float value;
	int flag = ([magnifyView flags] & NX_ALTERNATEMASK);
	
	value = [(slider = (flag) ? gridSlider : scaleSlider) floatValue] + [[sender selectedCell] tag];
	
	[slider setFloatValue:MAX(MIN(value, [(Slider *) slider maxValue]), [(Slider *) slider minValue])];
	
	return (flag) ? [self changeGrid:slider] : [self changeScale:slider];
}

- changeScale:sender
{
	float scale, value = rint([sender floatValue]); /* 'intValue' truncates */

	scale = (value < -1.0) ? 1.0 / (- value) : 2.0 + value; /* non-linear slider */
	value = floor(scale * 10.0) / 10.0; /* round to nearest .1 */
	
	[sender setTag:YES];

	return [self setScale:(NXCoord) value];
}

- setScale:(float) scale
{
	char title[] = WINDOWTITLEPROTOTYPE;
	float value = rint((scale < 1) ? - (1 / scale) : scale - 2.0);

	[[magnifyView setScale:(NXCoord) scale] createWindows];

	if ([scaleSlider floatValue] != value) [scaleSlider setFloatValue:value];
	if ([scaleField floatValue] != scale) [scaleField setFloatValue:scale];

	(void) sprintf(title, WINDOWTITLEFORMAT, scale);
	[magnifyWindow setTitle:title];

	return self;
}

- changeGrid:sender
{
	float size = rint([gridSlider floatValue]);

	[sender setTag:YES];

	return [self setGridSize:(NXCoord) size andEnable:(BOOL) [gridButton intValue]];
}

- setGridSize:(float) size andEnable:(BOOL) flag
{
	[[magnifyView setGridSize:(NXCoord) size] setShowGrid:flag];

	[[gridField setFloatValue:size] setEnabled:[magnifyView showGrid]];

	if ([gridSlider floatValue] != size) [gridSlider setFloatValue:size];

	return self;
}

- changeUpdate:sender
{
	double value = rint([updateSlider floatValue] * 120.0) / 120.0; /* round to nearest 0.01 */

	[sender setTag:YES];

	return [self setUpdateRate:value andEnable:[updateButton intValue]];
}

- setUpdateRate:(double) update andEnable:(BOOL) flag
{
	[[magnifyView setUpdateRate:update] setAutoUpdate:flag];

	[[updateField setFloatValue:update] setEnabled:[magnifyView autoUpdate]];
		
	return self;
}

- toggleFrozen:sender
{
	[magnifyView setFrozen:![magnifyView isFrozen]];
	
	[[self mainMenu] update];
	
	return self;
}

- toggleCursor:sender
{
	id cursor = nil;
	int tag = [[sender selectedCell] tag];

	switch (tag) {
		case CROSS : cursor = crossCursor; break;
		case CURRENT : cursor = nil; break;
		default: cursor = nil;
		}
	
	[[magnifyView setFakeCursor:cursor] setShowFakeCursor:(tag != NONE)];
	
	[sender setTag:YES];

	return self;
}

- magnifyView { return magnifyView; }

- loadDefaultSets
{
	char *r, *s, *scratch = NXCopyStringBuffer(getDefault("DefaultSets"));
	
	for (r = s = strtok(scratch, SEPARATOR); s != NULL; s = strtok(NULL, SEPARATOR)) {
		if ([defaultPopUpList indexOfItem:s] == -1) [defaultPopUpList addItem:s];
		}

	[self selectDefaultSetNamed:r];

	NX_FREE(scratch);
	
	return self;
}

- saveDefaultSetList
{
	char buffer[BUFSIZ];
	unsigned int i, count;
	const char *title, *selected = [defaultButton title];

	if (![saveButton tag] && ![removeButton tag]) return self;

	(void) strcpy(buffer, selected);
	
	for (i = 0, count = [defaultPopUpList count]; i < count; i++) {
		if (!streq(selected, (title = [defaultPopUpList itemAt:i]))) (void) strcat(strcat(buffer, ":"), title);
		}
	
	writeDefault("DefaultSets", buffer);
	
	return self;
}

- selectDefaultSet:sender { return [self selectDefaultSetNamed:[[sender selectedCell] title]]; }

- selectDefaultSetNamed:(const char *) name
{
	int tag;
	const char *frame;
	BOOL flag1, flag2;
	double scale, grid, update;

	if (!streq([defaultButton title], name)) [defaultButton setTitle:name];
	if (!streq([defaultField stringValue], name)) [defaultField setStringValue:name];
	if ([defaultPopUpList indexOfItem:name] == -1) [defaultPopUpList insertItem:name at:0];
	
	if ((flag1 = streq(name, ORIGINALDEFAULTSET))) name = NULL;
	[removeButton setEnabled:!flag1];
	
	scale = getFloatDefaultSet(name, "Scale");
	grid = getFloatDefaultSet(name, "GridSize");
	[gridButton setIntValue:(flag1 = getBoolDefaultSet(name, "ShowGrid"))];

	update = getFloatDefaultSet(name, "UpdateInterval");
	[updateButton setIntValue:(flag2 = getBoolDefaultSet(name, "AutoUpdate"))];
	[updateSlider setFloatValue:update];

	if ((frame = getDefaultSet(name, "NXWindow Frame " WINDOWNAME)) != NULL) (void) [magnifyWindow setFrameFromString:frame];

	[cursorButton selectCellWithTag:getIntDefaultSet(name, "Cursor")];
	tag = [cursorButton tag];
	[self toggleCursor:cursorButton];
	[cursorButton setTag:tag]; /* undo side effect of toggleCursor: */

	return [[[self setScale:scale] setGridSize:grid andEnable:flag1] setUpdateRate:update andEnable:flag2];
}

- saveDefaultSet:sender 
{
	char *string = NXCopyStringBuffer([defaultField stringValue]);
	
	[saveButton setTag:YES];

	[self saveDefaultSetNamed:string]; /* have to use copy of string as contents may change */

	NX_FREE(string);

	return self; 
}

- saveDefaultSetNamed:(const char *) name
{
	const char *key = name;

	if (name == NULL || *name == '\0') return self;
	else if (streq(key, ORIGINALDEFAULTSET)) key = NULL;

	if ([scaleSlider tag]) (void) writeDefaultSet(key, "Scale", [scaleField stringValue]);
	
	if ([gridButton tag]) (void) writeBoolDefaultSet(key, "ShowGrid", [gridButton intValue]);
	if ([gridSlider tag]) (void) writeDefaultSet(key, "GridSize", [gridField stringValue]);
	
	if ([updateButton tag]) (void) writeBoolDefaultSet(key, "AutoUpdate", [updateButton intValue]);
	if ([updateSlider tag]) (void) writeDefaultSet(key, "UpdateInterval", [updateField stringValue]);
	
	if ([cursorButton tag]) (void) writeIntDefaultSet(key, "Cursor", [[cursorButton selectedCell] tag]);

	if ([magnifyView resized]) {
		char frame[NX_MAXFRAMESTRINGLENGTH];

		(void) [magnifyWindow saveFrameToString:frame]; /* saveFrameToString: should return (id) 'self', not (void)! */
		writeDefaultSet(key, "NXWindow Frame " WINDOWNAME, frame);
		}

	return [[self selectDefaultSetNamed:name] saveDefaultSetList];
}

- removeDefaultSet:sender
{
	char *string = NXCopyStringBuffer([defaultField stringValue]);
	
	[removeButton setTag:YES];

	[self removeDefaultSetNamed:string]; /* have to use copy of string as contents may change */

	NX_FREE(string);

	return self; 
}

- removeDefaultSetNamed:(const char *) name
{
	if (*name == '\0' || streq(name, ORIGINALDEFAULTSET)) return nil;

	(void) removeDefaultSet(name, "Scale");
	(void) removeDefaultSet(name, "ShowGrid");
	(void) removeDefaultSet(name, "GridSize");
	(void) removeDefaultSet(name, "AutoUpdate");
	(void) removeDefaultSet(name, "UpdateInterval");
	(void) removeDefaultSet(name, "Cursor");
	(void) removeDefaultSet(name, "NXWindow Frame " WINDOWNAME);

	if (streq([defaultButton title], name)) [self selectDefaultSetNamed:ORIGINALDEFAULTSET];

	(void) [defaultPopUpList removeItem:name];
	
	return [self saveDefaultSetList];
}

- setDefaultPopUpList:object
{
	[[(defaultPopUpList = object) setTarget:self] setAction:@selector(selectDefaultSet:)];
	
	return self;
}

- setVersion:anObject
{
	[(version = anObject) setStringValue:VERSION];
	
	return self;
}

- (int) msgVersion:(const char **) aString ok:(int *) flag
{
	*flag = ((*aString = [version stringValue]) != NULL);

	return SEND_SUCCESS;
}

- (BOOL) isFrozenUpdateAction:(MenuCell *) anObject
{
	BOOL flag = [[NXApp magnifyView] isFrozen];
	
	if ([anObject tag] == flag) return NO;
	
	[[anObject setTag:(int) flag] setTitle:(flag) ? UNFROZENMENUTITLE : FROZENMENUTITLE];
	[anObject setKeyEquivalent:tolower([anObject title][0])];
	
	return YES;
}

@end
