/* File: MagnifyView.m - View class for 'Magnify'
 *
 * By: Christopher Lane <lane@camis.stanford.edu>
 * Symbolic Systems Resources Group
 * Knowledge Systems Laboratory
 * Stanford University
 *
 * PostScript optimizations by: Scott Byer <byer@adobe.com>
 *
 * Date:  13 May 1994
 *
 * Copyright: 1990, 1991, 1992  & 1994 by The Leland Stanford Junior University.
 * This program may be distributed without restriction for non-commercial use.
 */
 
#import "Magnify.h"
#import "MagnifyView.h"
#import "MagnifyWraps.h"
#import "NXCursorPatch.h"

#define HALF ((NXCoord) 0.5)

#define NX_NOBUTTONS (0)
#define NX_EMPTYMASK (NX_COMMANDMASK & (~ NX_COMMANDMASK))

#define MAGNIFYLEVEL (100)
#define NXSetWindowLevel _NXSetWindowLevel

extern int NXSetWindowLevel(int, int);
extern void timer(DPSTimedEntry, double, MagnifyView *);

@implementation MagnifyView

- initFrame:(const NXRect *) frameRect;
{
	[super initFrame:frameRect];

	flags = NX_EMPTYMASK;
	mouse = frameRect->origin;
	
	gridSize = scale = (NXCoord) 4.0;
	fakeCursor = nil;
	updateRate = 0.1;
	showFakeCursor = autoUpdate = showGrid = resized = frozen = NO;

	usedGridNumArray = lenGridNumArray = 0;

	invisibleImage = [[NXImage alloc] setSize:&bounds.size];

	return [[self resizeGrid] createWindows];
}

- (BOOL) acceptsFirstResponder { return YES; }

- copy:sender
{
	char *buffer;
	NXStream *stream;
	int length, maxLength;
	Pasteboard *pasteboard;

	if ((stream = NXOpenMemory(NULL, 0, NX_WRITEONLY)) == NULL) return nil;

	[(pasteboard = [Pasteboard new]) declareTypes:&NXPostScriptPboardType num:1 owner:self];

	[self copyPSCodeInside:&bounds to:stream];
	NXFlush(stream);

	NXGetMemoryBuffer(stream, &buffer, &length, &maxLength);
	[pasteboard writeType:NXPostScriptPboard data:buffer length:length];
	NXCloseMemory(stream, NX_FREEBUFFER);

	return self;
}

- drawSelf:(const NXRect *) rects :(int) rectCount
{
	static NXPoint origin = {0, 0};

	[invisibleImage composite:NX_COPY toPoint:&origin];
	
	return self;
}

- free
{
	[self resetTimedEntry];
	
	[invisibleImage free];

	if (gridNumArray != NULL) free((void *) gridNumArray);
	if (invisibleWindow != nil) [invisibleWindow close];

	return [super free];
}

- sizeTo:(NXCoord) newWidth :(NXCoord) newHeight;
{	
	id result = [super sizeTo:newWidth :newHeight];
	
	[[self resizeGrid] createWindows];
	
	[invisibleImage setSize:&bounds.size];
	
	resized = YES;
	
	return result;
}

- mouseMoved:(NXEvent *) theEvent
{
	NXEvent toss;
	id result = [nextResponder mouseMoved:theEvent];

	if ([NXApp peekNextEvent:NX_MOUSEMOVEDMASK into:&toss] == NULL) {
		mouse = theEvent->location;
		[window convertBaseToScreen:&mouse];
		[[self updateImage] display];
		}

	return result;
}

- flagsChanged:(NXEvent *) theEvent
{
	flags = theEvent->flags;

	return [nextResponder flagsChanged:theEvent];
}

- updateImage
{
	(void) [invisibleImage lockFocus]; {
		MVWimage(mouse.x - offset.width, mouse.y - offset.height);
		if (showGrid) [self drawGrid];
		if (showFakeCursor) [self drawFakeCursor];
		} [invisibleImage unlockFocus];
				
	return self;
}

- drawFakeCursor
{
	NXSize size;
	NXPoint point; 
	NXImage *image;
	const NXPoint *hotSpot;
	NXCursor *cursor = fakeCursor;
	
	if (cursor == nil) if ((cursor = [NXCursor currentCursor]) == nil) return nil;
	
	[(image = [cursor image]) getSize:&size];

	hotSpot = [cursor hotSpot];

	point.x = (offset.width * scale) + ((scale > 1) ? (scale * HALF) : 0) - hotSpot->x;
	point.y = (offset.height * scale) + ((scale > 1) ? (scale * HALF) : 0) - (size.height + hotSpot->y);

	[image composite:NX_SOVER toPoint:&point];

	return self;
}

- drawGrid
{	
	if (gridNumArray != NULL) MVWdrawgrid(gridNumArray, usedGridNumArray);

	return self;
}

- resizeGrid
{
	double x, y, max;
	unsigned int entriesNeeded;

	/* Check the space required for the numbers.  Each rectangle requires four floats - x, y, width, height. */
	entriesNeeded = ceil(bounds.size.height / (double) gridSize) + ceil(bounds.size.width / (double) gridSize);
	entriesNeeded *= 4;

	if (entriesNeeded > lenGridNumArray) {
		lenGridNumArray = entriesNeeded;
		if (gridNumArray == NULL) gridNumArray = (NXCoord *) malloc(sizeof(NXCoord) * entriesNeeded);
		else gridNumArray = (NXCoord *) realloc(gridNumArray, (sizeof(NXCoord) * entriesNeeded));
		if (gridNumArray == NULL) {
			lenGridNumArray = 0;
			return self;
			}
		}

	entriesNeeded = 0; /* Zero out entriesNeeded so that it can be reused as a counter. */

	max = bounds.origin.y + bounds.size.height;
	for (y = bounds.origin.y; y < max; y += gridSize) {
		gridNumArray[entriesNeeded++] = bounds.origin.x;
		gridNumArray[entriesNeeded++] = y;
		gridNumArray[entriesNeeded++] = bounds.size.width;
		gridNumArray[entriesNeeded++] = 0;
		}

	max = bounds.origin.x + bounds.size.width;
	for (x = bounds.origin.x; x < max; x += gridSize) {
		gridNumArray[entriesNeeded++] = x;
		gridNumArray[entriesNeeded++] = bounds.origin.y;
		gridNumArray[entriesNeeded++] = 0;
		gridNumArray[entriesNeeded++] = bounds.size.height;
		}

	usedGridNumArray = entriesNeeded; /* Set the value of the number we are going to use. */

	return self;
}

- (int) flags { return flags; }

- (NXPoint *) mouse { return &mouse; }

- (NXCoord) scale { return scale; }
- setScale:(NXCoord) newScale; { scale = newScale; return self; }

- (NXCoord) gridSize { return gridSize; }

- setGridSize:(NXCoord) value
{
	gridSize = value;

	return [[[self resizeGrid] updateImage] display];
}

- (float) updateRate { return updateRate; }

- setUpdateRate:(float) value
{
	BOOL flag = (updateRate != value);

	updateRate = value;

	if (autoUpdate & flag) [self setupTimedEntry];

	return self;
}

- (BOOL) autoUpdate { return autoUpdate; }

- setAutoUpdate:(BOOL) value
{
	SEL action = ((autoUpdate = value) == YES) ? @selector(setupTimedEntry) : @selector(resetTimedEntry);
	
	[self perform:action with:self afterDelay:0 cancelPrevious:YES]; /* delay fixes subtle interface timing bug */

	return self;
}

- setupTimedEntry
{
	if (timedEntry != NULL) {
		DPSRemoveTimedEntry(timedEntry);
		timedEntry = NULL;
		}

	if ((timedEntry = DPSAddTimedEntry(updateRate, (DPSTimedEntryProc) &timer, (void *) self, NX_BASETHRESHOLD)) != NULL) {
		(void) [window removeFromEventMask:NX_MOUSEMOVEDMASK];
		}

	return self;
}

- resetTimedEntry
{
	if (timedEntry != NULL) {
		DPSRemoveTimedEntry(timedEntry);
		timedEntry = NULL;
		}
	
	if (!frozen) (void) [window addToEventMask:NX_MOUSEMOVEDMASK];

	return [[self updateImage] display];
}

- (BOOL) resized { return resized; }

- (BOOL) isFrozen { return frozen; }

- setFrozen:(BOOL) state
{
	if ((frozen = state) == YES) [self freeze];
	else [self unfreeze];
	
	return self;
}

- freeze
{
	if (autoUpdate) [self resetTimedEntry];
	(void) [window removeFromEventMask:NX_MOUSEMOVEDMASK]; /* undo side effect of resetTimedEntry */

	return self;
}

- unfreeze
{
	if (!frozen) [self setAutoUpdate:autoUpdate];

	return self;
}

- (BOOL) showGrid { return showGrid; }
- setShowGrid:(BOOL) state { showGrid = state; return [[self updateImage] display]; }

- fakeCursor { return fakeCursor; }
- setFakeCursor:cursor { fakeCursor = cursor; return self; }

- (BOOL) showFakeCursor { return showFakeCursor; }
- setShowFakeCursor:(BOOL) state { showFakeCursor = state; return [[self updateImage] display]; }

- createWindows
{
	NXRect rect = bounds;
	int pixelsWide, pixelsHigh;
		
	rect.size.width = ceil(bounds.size.width / scale);
	offset.width = rint(bounds.size.width / scale * HALF);

	rect.size.height = ceil(bounds.size.height / scale);
	offset.height = rint(bounds.size.height / scale * HALF);

	if (invisibleWindow == nil) {
		invisibleWindow = [[Window alloc] initContent:&rect style:NX_PLAINSTYLE backing:NX_NONRETAINED buttonMask:NX_NOBUTTONS defer:NO];
		PSsetautofill(NO, [invisibleWindow windowNum]);
		NXSetWindowLevel([invisibleWindow windowNum], MAGNIFYLEVEL);
		[(invisibleView = [invisibleWindow contentView]) allocateGState];
		}
	else [[invisibleWindow sizeWindow:rect.size.width :rect.size.height] moveTo:-rect.size.width :0];

	[invisibleView getFrame:&rect];

	[[[invisibleWindow disableDisplay] disableCursorRects] setAvoidsActivation:YES];
	(void) [invisibleWindow setEventMask:NX_NULLEVENTMASK];

	pixelsWide = rect.size.width;
	pixelsHigh = rect.size.height;
	
	(void) [invisibleView lockFocus]; { /* Store away gstate which has window as the context's windowdevice. */
		MVWinvisibleGState(scale, pixelsWide, pixelsHigh);
		} [invisibleView unlockFocus];

	return [[self updateImage] display];
}

void timer(DPSTimedEntry teNumber, double now, MagnifyView *self)
{
	if (![self isFrozen]) {
		NXPoint *mouse = [self mouse];
		Window *window = [self window];
		
		[window getMouseLocation:mouse];
		[window convertBaseToScreen:mouse];
		[[self updateImage] display];
		}
}

@end
